{% if ip_ver == 4 %}
{% set max_pref_length = 32 %}
{% else %}
{% set max_pref_length = 128 %}
{% endif %}

{% include "header.j2" %}


{{ "header"|include_local_file -}}
{% if ip_ver == 4 %}
{{ "header4"|include_local_file -}}
{% else %}
{{ "header6"|include_local_file -}}
{% endif %}


{% include "rpki.j2" %}


{% include "common.j2" %}


{% include "irrdb.j2" %}


{% include "clients.j2" %}
 

{{ "footer"|include_local_file -}}
{% if ip_ver == 4 %}
{{ "footer4"|include_local_file -}}
{% else %}
{{ "footer6"|include_local_file -}}
{% endif %}
